import os
import json
import sys

from .env import CortexEnv

TOKEN_ENV_VAR_NAME_FOR_DAEMONS = "startup.token"
API_ENDPOINT_FOR_DAEMONS = "http://cortex-kong.cortex.svc.cluster.local:8000"


# -------------------------------------- Token Loaders --------------------------------------
def load_token_from_job_env():
    try:
        return json.loads(sys.argv[1]).get("token", None)
    except Exception:
        return None


def load_token_from_daemon_env():
    return os.getenv(TOKEN_ENV_VAR_NAME_FOR_DAEMONS, None)


def load_token_from_local_env():
    return CortexEnv().token


def load_token(env_resolution_order=["args","local-cli","local-env","job","daemon"], token=None):
    """
    Returns the first token it can find based on the resolution order
    """
    token_finders_per_env = {
        "args": (lambda: token),
        "daemon": load_token_from_daemon_env,
        "job": load_token_from_job_env,
        "local-env": CortexEnv.get_cortex_token,
        "local-cli": (lambda: CortexEnv.get_cortex_profile().get("token", None))
    }
    for env in env_resolution_order:
        if not env in token_finders_per_env:
            raise Exception(f"Invalid Env: {env}")
        else:
            _token = token_finders_per_env[env]()
        if _token is not None:
            return _token
    # raise Exception(f"Could not find token in envs: {env_resolution_order}")
    return None

# ------------------------------------ Endpoint Loaders ------------------------------------


def load_endpoint_from_daemon_env():
    return API_ENDPOINT_FOR_DAEMONS


def load_endpoint_from_job_env():
    try:
        return json.loads(sys.argv[1]).get("apiEndpoint", None)
    except Exception:
        return None


def load_api_endpoint(env_resolution_order=["args","local-cli","local-env","job","daemon"], endpoint=None):
    """
    Returns the first api endpoint it can extract from the enviroment.
    The order of in which the environments are searched is dictated by the env_resolution_order
    """
    endpoint_finders_per_env = {
        "args": (lambda: endpoint),
        "daemon": load_endpoint_from_daemon_env,
        "job": load_endpoint_from_job_env,
        "local-env": (lambda: os.getenv('CORTEX_URI', None)),
        "local-cli": (lambda: CortexEnv.get_cortex_profile().get("url", None))
    }
    for env in env_resolution_order:
        if not env in endpoint_finders_per_env:
            raise Exception(f"Invalid Env: {env}")
        else:
            _endpoint = endpoint_finders_per_env[env]()
        if _endpoint is not None:
            return _endpoint
    # raise Exception(f"Could not find endpoint in envs: {env_resolution_order}")
    return None


# Developing Mongo Persist Handler
import pandas as pd
from aistac.handlers.abstract_handlers import AbstractSourceHandler, AbstractPersistHandler
from aistac.handlers.abstract_handlers import HandlerFactory, ConnectorContract

__author__ = 'Darryl Oatridge, Omar Eid'


class MongoSourceHandler(AbstractSourceHandler):
    """ A mongoDB source handler"""

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Hander passing the source_contract dictionary """
        # required module import
        self.mongo = HandlerFactory.get_module('pymongo')
        super().__init__(connector_contract)
        database = self.connector_contract.kwargs.get("database")
        self._mongo_database = self.mongo.MongoClient(self.connector_contract.uri)[database]
        self._mongo_collection = self._mongo_database[self.connector_contract.kwargs.get("collection")]
        self._file_state = 0
        self._changed_flag = True

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['mongo']

    def load_canonical(self, **kwargs) -> pd.DataFrame:
        """ returns the canonical dataset based on the source contract
            The canonical in this instance is a dictionary that has the headers as the key and then
            the ordered list of values for that header
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The PandasSource Connector Contract has not been set")

        _cc = self.connector_contract

        load_params = _cc.kwargs
        load_params.update(_cc.query)  # Update kwargs with those in the uri query
        load_params.update(kwargs)     # Update with any passed though the call
        if load_params.get("aggregate") is not None:
            return pd.DataFrame(list(self._mongo_collection.aggregate(_cc.kwargs.get("aggregate", []))))
        elif load_params.get("query") is not None:
            base_query = self._mongo_collection.find(_cc.kwargs.get("find", load_params.get("query", {})),
                                                     _cc.kwargs.get("project"))
            if _cc.kwargs.get("limit") is not None:
                base_query.limit(_cc.kwargs.get("limit"))
            if _cc.kwargs.get("skip") is not None:
                base_query.skip(_cc.kwargs.get("skip"))
            if _cc.kwargs.get("sort") is not None:
                base_query.sort(_cc.kwargs.get("sort"))
            return pd.DataFrame(list(base_query))
        return pd.DataFrame()

    def exists(self) -> bool:
        """ returns True if the collection exists """
        _cc = self.connector_contract
        return _cc.kwargs.get("collection") in self._mongo_database.list_collection_names()

    def has_changed(self) -> bool:
        """ returns the amount of documents in the collection
            ... if the counts change ... then the collection was probably modified ...
            ... this assumes that records are never edited/updated ... nor deleted ...
        """
        _cc = self.connector_contract
        state = self._mongo_collection.count_documents(_cc.kwargs.get("find", {}))
        if state != self._file_state:
            self._changed_flag = True
            self._file_state = state
        return self._changed_flag

    def reset_changed(self, changed: bool = False):
        """ manual reset to say the file has been seen. This is automatically called if the file is loaded"""
        changed = changed if isinstance(changed, bool) else False
        self._changed_flag = changed


class MongoPersistHandler(MongoSourceHandler, AbstractPersistHandler):
    # a mongoDB persist handler

    def persist_canonical(self, canonical: pd.DataFrame, **kwargs) -> bool:
        """ persists the canonical dataset
        Extra Parameters in the ConnectorContract kwargs:
            - file_type: (optional) the type of the source file. if not set, inferred from the file extension
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _uri = self.connector_contract.uri
        return self.backup_canonical(uri=_uri, canonical=canonical, **kwargs)

    def backup_canonical(self, canonical: pd.DataFrame, uri: str, **kwargs) -> bool:
        """ creates a backup of the canonical to an alternative URI  """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _cc = self.connector_contract
        _address = _cc.parse_address(uri=uri)
        persist_params = kwargs if isinstance(kwargs, dict) else _cc.kwargs
        persist_params.update(_cc.parse_query(uri=uri))

        if _cc.kwargs.get("ordered") is None:
            resp = self._mongo_collection.insert_many(canonical.to_dict(orient="records"))
        else:
            resp = self._mongo_collection.insert_many(canonical.to_dict(orient="records"),
                                                      ordered=_cc.kwargs.get("ordered"))
        # print(dir(resp))
        return len(resp.inserted_ids) == canonical.shape[0]

    def remove_canonical(self) -> bool:
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _cc = self.connector_contract
        raise NotImplementedError("remove_canonical for MongoPersistHandler not yet implemented.")



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import logging
from requests.exceptions import HTTPError
from pathlib import Path


def log_message(msg: str, log: logging.Logger, level=logging.INFO, *args, **kwargs):
    """
    Logs a message.

    :param msg: Message to log
    :param log: logger where message should be logged
    :param level: optional log level, defaults to INFO
    :param args: a tuple of arguments passed to the logger
    :param kwargs: a dictionary of keyword arguments passed to the logger
    """
    log.log(level, msg, *args, **kwargs)


def get_cortex_profile(profile_name=None):
    """
    Gets the current cortex profile or the profile that matches the optionaly given name.
    """
    cortex_config_path = Path.home() / '.cortex/config'

    if cortex_config_path.exists():
        with cortex_config_path.open() as f:
            cortex_config = json.load(f)

        if profile_name is None:
            profile_name = cortex_config.get('currentProfile')

        return cortex_config.get('profiles', {}).get(profile_name, {})
    return {}


def get_logger(name):
    """
    Gets a logger with the given name.
    """
    log = logging.getLogger(name)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s/%(module)s: %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    log.addHandler(handler)
    log.setLevel(logging.INFO)
    return log


def raise_for_status_with_detail(resp):
    """
    wrap raise_for_status and attempt give detailed reason for api failure
    re-raise HTTPError for normal flow
    :param resp: python request resp
    :return:
    """
    try:
        resp.raise_for_status()
    except HTTPError as http_exception:
        try:
            log_message(msg=resp.json(), log=get_logger('http_status'), level=logging.ERROR)
        except Exception as e:
            pass # resp.json() failed
        finally:
            raise http_exception

